# This file is part of scrilla: https://github.com/chinchalinchin/scrilla.

# scrilla is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.

# scrilla is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with scrilla.  If not, see <https://www.gnu.org/licenses/>
# or <https://github.com/chinchalinchin/scrilla/blob/develop/main/LICENSE>.

"""
This module interfaces with the external services the program uses to hydrate with financial data. In the case of price and interest history, the functions in this module defer to the cache before making expensive HTTP requests. Statistical data retrieved from FRED and dividment payment histories are not persisted in the cache since most of the data is reported on an irregular basis and it is impossible to tell based on the date alone whether or not the cache is out of date.

This module can be imported and used directly in other Python scripts if the API keys for the services have been set.

```
import os
from scrilla.services import get_daily_price_history

os.environ.setdefault('ALPHA_VANTAGE_KEY')
prices = get_daily_price_history('AAPL')
```
"""
import itertools
import time
import requests
from typing import Dict, List, Union

from datetime import date

from scrilla import settings, errors, cache
from scrilla.static import keys, constants
from scrilla.util import outputter, helper, dater

logger = outputter.Logger("scrilla.services", settings.LOG_LEVEL)


class StatManager():
    """
    StatManager is an interface between the application and the external services that hydrate it with financial statistics data. This class gets instantiated on the level of the `scrilla.services` module with the value defined in `scrilla.settings.STAT_MANAGER`. This value is in turn defined by the value of the `STAT_MANAGER` environment variable. This value determines how the url is constructed, which API credentials get appended to the external query and the keys used to parse the response JSON container the statistical data.

    Attributes
    ----------
    1. **genre**: ``str``
        A string denoting which service will be used for data hydration. Genres can be accessed through the `keys.keys['SERVICES']` dictionary.
    2. **self.service_map**: ``dict``
        A dictionary containing keys unique to the service defined by `genre`, such as endpoints, query parameters, etc. 

    Raises
    ------
    1. **scrilla.errors.ConfigurationError**
        If the **STAT_MANAGER** environment variable has been set to a value the program doesn't understand or hasn't set at all, this error will be thrown when this class is instantiated.

    .. notes ::
        * This class handles retrieval of financial statistics from the [Quandl's FRED dataset](https://data.nasdaq.com/data/FRED-federal-reserve-economic-data) and [Quandl's USTREASURY dataset](https://data.nasdaq.com/data/USTREASURY-us-treasury). In other words, it handles statistics and interest rate service calls. Interest rates are technically prices, not statistics, but...well, I don't have a good explanation why this isn't `scrilla.services.PriceManager`; I suppose I grouped these classes more by service type than data type. Might do to refactor...
    """

    def __init__(self, genre):
        self.genre = genre
        self.service_map = None
        if self._is_quandl():
            self.service_map = keys.keys["SERVICES"]["STATISTICS"]["QUANDL"]["MAP"]
            self.key = settings.Q_KEY
            self.url = settings.Q_URL
        if self.service_map is None:
            raise errors.ConfigurationError(
                'No STAT_MANAGER found in the parsed environment settings')

    def _is_quandl(self):
        """
        Returns
        -------
        `bool`
            `True` if this instace of `StatManager` is a Quandl interface. `False` otherwise.

        .. notes::
            * This is for use within the class and probably won't need to be accessed outside of it. `StatManager` is intended to hide the data implementation from the rest of the library, i.e. it is ultimately agnostic about where the data comes where. It should never need to know `StatManger` is a Quandl interface. Just in case the library ever needs to populate its data from another source.

        """
        if self.genre == keys.keys['SERVICES']['STATISTICS']['QUANDL']['MANAGER']:
            return True
        return False

    def _construct_query(self, start_date: date, end_date: date) -> str:
        """
        Constructs and formats the query parameters for the external statistics service. Note, this method appends the API key to the query. Be careful with the returned value.

        Parameters
        ----------
        1. **start_date** : ``datetime.date``
            Start date of historical sample to be retrieved.
        2. **end_date** : ``datetime.date``
            End date of historical sample to be retrieved.

        Returns
        -------
        ``str``
            The formatted query for the specific service defined by `self.genre`.

        """
        query = ""
        if end_date is not None:
            end_string = dater.to_string(end_date)
            query += f'&{self.service_map["PARAMS"]["END"]}={end_string}'

        if start_date is not None:
            start_string = dater.to_string(start_date)
            query += f'&{self.service_map["PARAMS"]["START"]}={start_string}'

        if query:
            logger.debug(f'StatManager Query (w/o key) = {query}')
            return f'{query}&{self.service_map["PARAMS"]["KEY"]}={self.key}'

        return f'{self.service_map["PARAMS"]["KEY"]}={self.key}'

    def _construct_stat_url(self, symbol: str, start_date: date, end_date: date):
        """
        Constructs the full URL path for the external statistics service. Note, this method will return the URL with an API key appended as a query parameter. Be careful with the returned value.

        Parameters
        ----------
        1. **symbol**: ``str``
            Symbol representing the statistical series to be retrieved. List of allowable symbols can be found [here](https://data.nasdaq.com/data/FRED-federal-reserve-economic-data)
        2. **start_date**: ``datetime.date`` 
            Start date of historical sample to be retrieved.
        3. **end_date**: ``datetime.date``
            End date of historical sample to be retrieved.


        Returns
        -------
        ``str``
            The formatted URL for the specific statistics service defined by `self.genre`.
        """
        url = f'{self.url}/{self.service_map["PATHS"]["FRED"]}/{symbol}?'
        url += self._construct_query(start_date=start_date, end_date=end_date)
        return url

    def _construct_interest_url(self, start_date, end_date):
        """
        Constructs the full URL path for the external interest rate service. Note, this method will return the URL with an API key appended as a query parameter. Be careful with the returned value.

        Parameters
        ----------
        1. **start_date**: ``datetime.date``
            Start date of historical sample to be retrieved.
        2. **end_date**: ``datetime.date`` 
            End date of historical sample to be retrieved.

        Returns
        -------
        ``str``
            The formatted URL for the specific interest rate service defined by `self.genre`.

        .. notes::
            * The URL returned by this method will always contain a query for a historical range of US Treasury Yields, i.e. this method is specifically for queries involving the "Risk-Free" (right? right? *crickets*) Yield Curve. 
        """
        url = f'{self.url}/{self.service_map["PATHS"]["YIELD"]}?'
        url += self._construct_query(start_date=start_date, end_date=end_date)
        return url

    def get_stats(self, symbol, start_date, end_date):
        url = self._construct_stat_url(symbol, start_date, end_date)
        response = requests.get(url).json()

        raw_stat = response[self.service_map["KEYS"]["FIRST_LAYER"]
                            ][self.service_map["KEYS"]["SECOND_LAYER"]]
        formatted_stat = {}

        for stat in raw_stat:
            formatted_stat[stat[0]] = stat[1]
        return formatted_stat

    def get_interest_rates(self, start_date, end_date):
        url = self._construct_interest_url(
            start_date=start_date, end_date=end_date)
        response = requests.get(url).json()

        raw_interest = response[self.service_map["KEYS"]
                                ["FIRST_LAYER"]][self.service_map["KEYS"]["SECOND_LAYER"]]
        formatted_interest = {}
        for rate in raw_interest:
            formatted_interest[rate[0]] = rate[1:]
        return formatted_interest

    @staticmethod
    def format_for_maturity(maturity, results):
        try:
            maturity_key = keys.keys['YIELD_CURVE'].index(maturity)
        except KeyError:
            raise errors.InputValidationError(
                f'{maturity} is not a valid maturity for US Treasury Bonds')

        formatted_interest = {}
        for result in results:
            formatted_interest[result] = results[result][maturity_key]
        return formatted_interest


class DividendManager():
    """
    Attributes
    ----------
    1. **genre**: ``str``
        A string denoting which service will be used for data hydration. Genres can be accessed through the `keys.keys['SERVICES']` dictionary.
    2. **self.service_map**: ``dict``
        A dictionary containing keys unique to the service defined by `genre`, such as endpoints, query parameters, etc. 
    """

    def __init__(self, genre):
        self.genre = genre
        if self.genre == keys.keys['SERVICES']['DIVIDENDS']['IEX']['MANAGER']:
            self.service_map = keys.keys['SERVICES']['DIVIDENDS']['IEX']['MAP']
            self.key = settings.iex_key()
            self.url = settings.IEX_URL

        if self.service_map is None:
            raise errors.ConfigurationError(
                'No DIV_MANAGER found in the parsed environment settings')

    def _construct_url(self, ticker):
        query = f'{ticker}/{self.service_map["PATHS"]["DIV"]}/{self.service_map["PARAMS"]["FULL"]}'
        url = f'{self.url}/{query}?{self.service_map["PARAMS"]["KEY"]}={self.key}'
        logger.debug(f'DivManager Query (w/o key) = {query}')
        return url

    def get_dividends(self, ticker):
        url = self._construct_url(ticker)
        response = requests.get(url).json()
        formatted_response = {}

        for item in response:
            this_date = str(item[self.service_map['KEYS']['DATE']])
            div = item[self.service_map['KEYS']['AMOUNT']]
            formatted_response[this_date] = div

        return formatted_response


class PriceManager():
    """
    PriceManager is an interface between the application and the external services that hydrate it with price data. This class gets instantiated on the level of the `scrilla.services` module with the value defined in `scrilla.settings.PRICE_MANAGER` variable. This value is in turn configured by the value of the **PRICE_MANAGER** environment variable. This value determines how the url is constructed, which API credentials get appended to the external query and the keys used to parse the response JSON containing the price data.

    Raises
    ------
    1. **scrilla.errors.ConfigurationError**
        If the **PRICE_MANAGER** environment variable hasn't been set or set to a value the program doesn't understand, this error will be thrown when this class is instantiated.

    Attributes
    ----------
    1. **genre**: ``str``
        A string denoting which service will be used for data hydration. Genres can be accessed through the `keys.keys['SERVICES']` dictionary.
    2. **self.service_map**: ``dict``
        A dictionary containing keys unique to the service defined by `genre`, such as endpoints, query parameters, etc. 

    """

    def __init__(self, genre):
        self.genre = genre
        if self.genre == keys.keys['SERVICES']['PRICES']['ALPHA_VANTAGE']['MANAGER']:
            self.service_map = keys.keys['SERVICES']['PRICES']['ALPHA_VANTAGE']['MAP']
            self.url = settings.AV_URL
            self.key = settings.av_key()
        if self.service_map is None:
            raise errors.ConfigurationError(
                'No PRICE_MANAGER found in the parsed environment settings')

    def _construct_url(self, ticker, asset_type):
        """
        Constructs the service url with the query and parameters appended. 

        Parameters
        ----------
        1. **ticker**: ``str`
            Ticker symbol of the asset whose prices are being retrieved.
        2. **asset_type**: ``str``
            Asset type of the asset whose prices are being retrieved. Options are statically
            accessible in the `scrillla.static` module dictionary `scrilla.keys.keys['ASSETS']`.

        Returns
        -------
        `str`
            The URL with the authenticated query appended, i.e. with the service's API key injected into the parameters. Be careful not to expose the return value of this function!

        .. notes::
            * this function will probably need substantially refactored if another price service is ever incorporated, unless the price service selected can be abstracted into the same template set by `scrilla.statics.keys['SERVICES']['PRICES']['ALPHA_VANTAGE']['MAP']`.
        """

        query = f'{self.service_map["PARAMS"]["TICKER"]}={ticker}'

        if asset_type == keys.keys['ASSETS']['EQUITY']:
            query += f'&{self.service_map["PARAMS"]["FUNCTION"]}={self.service_map["ARGUMENTS"]["EQUITY_DAILY"]}'
            query += f'&{self.service_map["PARAMS"]["SIZE"]}={self.service_map["ARGUMENTS"]["FULL"]}'

        elif asset_type == keys.keys['ASSETS']['CRYPTO']:
            query += f'&{self.service_map["PARAMS"]["FUNCTION"]}={self.service_map["ARGUMENTS"]["CRYPTO_DAILY"]}'
            query += f'&{self.service_map["PARAMS"]["DENOMINATION"]}={constants.constants["DENOMINATION"]}'

        auth_query = query + f'&{self.service_map["PARAMS"]["KEY"]}={self.key}'
        url = f'{self.url}?{auth_query}'
        logger.debug(f'PriceManager query (w/o key) = {query}')
        return url

    def get_prices(self, ticker: str, start_date: date, end_date: date, asset_type: str):
        """
        Retrieve prices from external service.

        Parameters
        ----------
        1. **ticker** : ``str``
            Ticker symbol of the asset whose prices are being retrieved.
        2. **start_date**; ``str``
        3. **end_date**: ``str``
        4. **asset_type** : ``str``
            Asset type of the asset whose prices are being retrieved. Options are statically
            accessible in the `scrillla.static` module dictionary `scrilla.keys.keys['ASSETS']`.

        Returns
        -------
        ``Dict[str, Dict[str, float]]``
        ```
            prices = { 
                        'date': { 
                            'open': value, 
                            'close': value
                        },  
                        'date': { 
                            'open': value,
                            'close': value
                        }
            }
        ```
         Dictionary of prices with date as key, ordered from latest to earliest.

        Raises
        ------
        1. **scrilla.errors.ConfigurationError**
            If one of the settings is improperly configured or one of the environment variables was unable to be parsed from the environment, this error will be thrown.
        2. **scrilla.errors.APIResponseError**
            If the service from which data is being retrieved is down, the request has been rate limited or some otherwise anomalous event has taken place, this error will be thrown.
        """
        url = self._construct_url(ticker, asset_type)
        response = requests.get(url).json()

        first_element = helper.get_first_json_key(response)
        # end function is daily rate limit is reached
        if first_element == self.service_map['ERRORS']['RATE_LIMIT']:
            raise errors.APIResponseError(
                response[self.service_map['ERRORS']['RATE_LIMIT']])
            # check for bad response
        if first_element == self.service_map['ERRORS']['INVALID']:
            raise errors.APIResponseError(
                response[self.service_map['ERRORS']['INVALID']])

        # check and wait for API rate limit refresh
        first_pass, first_element = True, helper.get_first_json_key(response)

        while first_element == self.service_map['ERRORS']['RATE_THROTTLE']:
            if first_pass:
                logger.comment(
                    f'{self.genre} API rate limit per minute exceeded. Waiting...')
                first_pass = False
            else:
                logger.comment('Waiting...')

            time.sleep(constants.constants['BACKOFF_PERIOD'])
            response = requests.get(url).json()
            first_element = helper.get_first_json_key(response)

            if first_element == self.service_map['ERRORS']['INVALID']:
                raise errors.APIResponseError(
                    response[self.service_map['ERRORS']['INVALID']])

        prices = self._slice_prices(
            start_date=start_date, end_date=end_date, asset_type=asset_type, prices=response)
        format_prices = {}
        for this_date in prices:
            close_price = self._parse_price_from_date(prices=prices, this_date=this_date, asset_type=asset_type,
                                                      which_price=keys.keys['PRICES']['CLOSE'])
            open_price = self._parse_price_from_date(prices=prices, this_date=this_date, asset_type=asset_type,
                                                     which_price=keys.keys['PRICES']['OPEN'])
            format_prices[this_date] = {
                keys.keys['PRICES']['OPEN']: open_price, keys.keys['PRICES']['CLOSE']: close_price}
        return format_prices

    def _slice_prices(self, start_date: date, end_date: date, asset_type: str, prices: dict) -> dict:
        """
        Parses the raw response from the external price service into a format the program will understand.

        Parameters
        ----------
        1. **start_date** : ``datetime.date``
        2. **end_date** : ``datetime.date``
        3. **asset_type** : ``str``
            Required: Asset type of the asset whose prices are being retrieved. Options are statically
            accessible in the `scrillla.static` module dictionary `scrilla.keys.keys['ASSETS']`.
        4. **response** : ``dict``
            The full response from the price manager, i.e. the entire price history returned by the external service in charge of retrieving pricce histories, the result returned from `scrilla.services.PriceManager.get_prices`

        Returns
        -------
        ``dict``: `{ 'date': value, 'date': value, ...}`
            Dictionary of prices with date as key, ordered from latest to earliest.


        Raises
        ------
        1. **KeyError**
            If the inputted or validated dates do not exist in the price history, a KeyError will be thrown. This could be due to the equity not having enough price history, i.e. it started trading a month ago and doesn't have 100 days worth of prices yet, or some other anomalous event in an equity's history. 
        2. **scrilla.errors.ConfigurationError**
            If one of the settings is improperly configured or one of the environment variables was unable to be parsed from the environment, this error will be thrown.
        """

        # NOTE: only really needed for `alpha_vantage` responses so far, due to the fact AlphaVantage either returns everything or 100 days or prices.
        # shouldn't need to verify genre anyway, since using service_map and service_map should abstract the response away.
        if self.genre == keys.keys['SERVICES']['PRICES']['ALPHA_VANTAGE']['MANAGER']:

            # NOTE: Remember AlphaVantage is ordered current to earliest. END_INDEX is
            # actually the beginning of slice and START_INDEX is actually end of slice.
            start_string, end_string = dater.to_string(
                start_date), dater.to_string(end_date)
            if asset_type == keys.keys['ASSETS']['EQUITY']:
                response_map = self.service_map['KEYS']['EQUITY']['FIRST_LAYER']
            elif asset_type == keys.keys['ASSETS']['CRYPTO']:
                response_map = self.service_map['KEYS']['CRYPTO']['FIRST_LAYER']

            start_index = list(prices[response_map].keys()).index(start_string)
            end_index = list(prices[response_map].keys()).index(end_string)
            prices = dict(itertools.islice(
                prices[response_map].items(), end_index, start_index+1))
            return prices

        raise errors.ConfigurationError(
            'No PRICE_MANAGER found in the parsed environment settings')

    def _parse_price_from_date(self, prices: Dict[str, Dict[str, float]], this_date: date, asset_type: str, which_price: str) -> str:
        """
        Parameters
        ----------
        1. **prices** : ``Dict[str, Dict[str, float]]``
            List containing the AlphaVantage response with the first layer peeled off, i.e.
            no metadata, just the date and prices.
        2. **date**: `date``
            String of the date to be parsed. Note: this is not a datetime.date object. String
            must be formatted `YYYY-MM-DD`
        3. **asset_type**: ``str``
            String that specifies what type of asset price is being parsed. Options are statically
            accessible in the `scrillla.static` module dictionary `scrilla.keys.keys['ASSETS']`
        4. **which_price**: ``str``
            String that specifies which price is to be retrieved, the closing price or the opening prices. Options are statically accessible 

        Returns
        ------
        ``str``
            String containing the price on the specified date.

        Raises
        ------
        1. **KeyError**
            If the inputted or validated dates do not exist in the price history, a KeyError will be thrown. This could be due to the equity not having enough price history, i.e. it started trading a month ago and doesn't have 100 days worth of prices yet, or some other anomalous event in an equity's history. 
        2. **scrilla.errors.InputValidationError**
            If prices was unable to be grouped into a (crypto, equity)-asset class or the opening/closing price did not exist for whatever reason, this error will be thrown.
        """
        if asset_type == keys.keys['ASSETS']['EQUITY']:
            if which_price == keys.keys['PRICES']['CLOSE']:
                return prices[this_date][self.service_map['KEYS']['EQUITY']['CLOSE']]
            if which_price == keys.keys['PRICES']['OPEN']:
                return prices[this_date][self.service_map['KEYS']['EQUITY']['CLOSE']]

        elif asset_type == keys.keys['ASSETS']['CRYPTO']:
            if which_price == keys.keys['PRICES']['CLOSE']:
                return prices[this_date][self.service_map['KEYS']['CRYPTO']['CLOSE']]
            if which_price == keys.keys['PRICES']['OPEN']:
                return prices[this_date][self.service_map['KEYS']['CRYPTO']['OPEN']]

        raise errors.InputValidationError(
            f'Verify {asset_type}, {which_price} are allowable values')


price_manager = PriceManager(settings.PRICE_MANAGER)
stat_manager = StatManager(settings.STAT_MANAGER)
div_manager = DividendManager(settings.DIV_MANAGER)
price_cache = cache.PriceCache()
interest_cache = cache.InterestCache()


def get_daily_price_history(ticker: str, start_date: Union[None, date] = None, end_date: Union[None, date] = None, asset_type: Union[None, str] = None) -> Dict[str, Dict[str, float]]:
    """
    Wrapper around external service request for price data. Relies on an instance of `PriceManager` configured by `settings.PRICE_MANAGER` value, which in turn is configured by the `PRICE_MANAGER` environment variable, to hydrate with data. 

    Before deferring to the `PriceManager` and letting it call the external service, however, this function checks if response is in local cache. If the response is not in the cache, it will pass the request off to `PriceManager` and then save the response in the cache so subsequent calls to the function can bypass the service request. Used to prevent excessive external HTTP requests and improve the performance of the application. Other parts of the program should interface with the external price data services through this function to utilize the cache functionality.

    Parameters
    ----------
    1. **ticker** :  ``str``
        Ticker symbol corresponding to the price history to be retrieved.
    2. **start_date** : ``datetime.date`` 
        *Optional*. Start date of price history. Defaults to None. If `start_date is None`, the calculation is made as if the `start_date` were set to 100 trading days ago. If `scrilla.files.get_asset_type(ticker)==scrill.keys.keys['ASSETS']['CRYPTO']`, this includes weekends and holidays. If `scrilla.files.get_asset_type(ticker)==scrilla.keys.keys['ASSETS']['EQUITY']`, this excludes weekends and holidays.
    3. **end_date** : ``datetime.date``
        Optional End date of price history. Defaults to None. If `end_date is None`, the calculation is made as if the `end_date` were set to today. If `scrilla.files.get_asset_type(ticker)==scrill.keys.keys['ASSETS']['CRYPTO']`, this means today regardless. If `scrilla.files.get_asset_type(ticker)==scrilla.keys.keys['ASSETS']['EQUITY']`, this excludes weekends and holidays so that `end_date` is set to the previous business date. 
    4. **asset_type** : ``string``
        *Optional*. Asset type of the ticker whose history is to be retrieved. Used to prevent excessive calls to IO and list searching. `asset_type` is determined by comparing the ticker symbol `ticker` to a large static list of ticker symbols maintained in installation directory's /data/static/ subdirectory, which can slow the program down if the file is constantly accessed and lots of comparison are made against it. Once an `asset_type` is calculated, it is best to preserve it in the process environment somehow, so this function allows the value to be passed in. If no value is detected, it will make a call to the aforementioned directory and parse the file to determine to the `asset_type`. Asset types are statically accessible through the `scrilla.keys.keys['ASSETS']` dictionary.

    Returns
    ------
    ``Dict[str, Dict[str, float]]`` : Dictionary with date strings formatted `YYYY-MM-DD` as keys and a nested dictionary containing the 'open' and 'close' price as values. Ordered from latest to earliest, e.g., 
    ```
        { 
            'date' : 
                { 
                    'open': value, 
                    'close': value  
                }, 
            'date': 
                { 
                    'open' : value, 
                    'close' : value 
                }, 
            ... 
        }
    ```

    Raises
    ------
    1. **scrilla.errors.PriceError**
        If no sample prices can be retrieved, this error is thrown. 

    .. notes::
        * The default analysis period, if no `start_date` and `end_date` are specified, is determined by the *DEFAULT_ANALYSIS_PERIOD** variable in the `settings,py` file. The default value of this variable is 100.
    """
    asset_type = errors.validate_asset_type(ticker, asset_type)
    start_date, end_date = errors.validate_dates(
        start_date, end_date, asset_type)

    cached_prices = price_cache.filter_price_cache(
        ticker=ticker, start_date=start_date, end_date=end_date)

    if cached_prices is not None:
        if asset_type == keys.keys['ASSETS']['EQUITY']:
            logger.debug(
                f'Comparing {len(cached_prices)} = {dater.business_days_between(start_date, end_date)}')
        elif asset_type == keys.keys['ASSETS']['CRYPTO']:
            logger.debug(
                f'Comparing {len(cached_prices)} = {dater.days_between(start_date, end_date)}')

    # make sure the length of cache is equal to the length of the requested sample
    if cached_prices is not None and dater.to_string(end_date) in cached_prices.keys() and (
        (asset_type == keys.keys['ASSETS']['EQUITY']
            and (dater.business_days_between(start_date, end_date)) == len(cached_prices))
        or
        (asset_type == keys.keys['ASSETS']['CRYPTO']
            and (dater.days_between(start_date, end_date)) == len(cached_prices))
    ):
        # TODO: debug the crypto out of date check.
        return cached_prices

    if cached_prices is not None:
        logger.debug(
            f'Cached {ticker} prices are out of date, passing request off to external service')

    prices = price_manager.get_prices(
        ticker=ticker, start_date=start_date, end_date=end_date, asset_type=asset_type)

    if cached_prices is not None:
        new_prices = helper.complement_dict_keys(prices, cached_prices)
    else:
        new_prices = prices

    for this_date in new_prices:
        open_price = new_prices[this_date][keys.keys['PRICES']['OPEN']]
        close_price = new_prices[this_date][keys.keys['PRICES']['CLOSE']]
        price_cache.save_row(ticker=ticker, date=this_date,
                             open_price=open_price, close_price=close_price)

    if not prices:
        raise errors.PriceError(
            f'Prices could not be retrieved for {ticker}')

    return prices


def get_daily_price_latest(ticker: str, asset_type: Union[None, str] = None) -> float:
    """
    Returns the latest closing price for a given ticker symbol.

    Parameters
    ----------
    1. **ticker**: ``str``
        ticker symbol whose latest closing price is to be retrieved. \n \n
    2. **asset_type**: str``
        *Optional*. Asset type of the ticker whose history is to be retrieved. Will be calculated from the `ticker` symbol if not provided.
    """
    last_date = dater.this_date_or_last_trading_date()
    prices = get_daily_price_history(
        ticker=ticker, asset_type=asset_type, start_date=last_date, end_date=last_date)
    first_element = helper.get_first_json_key(prices)
    return prices[first_element][keys.keys['PRICES']['OPEN']]


def get_daily_prices_latest(tickers: List[str], asset_types: Union[None, List[str]] = None):
    return {ticker: get_daily_price_latest(ticker, asset_types[i]) for i, ticker in enumerate(tickers)}


def get_daily_fred_history(symbol: str, start_date: Union[date, None] = None, end_date: Union[date, None] = None) -> list:
    """
    Wrapper around external service request for financial statistics data constructed by the Federal Reserve Economic Data. Relies on an instance of `StatManager` configured by `settings.STAT_MANAGER` value, which in turn is configured by the `STAT_MANAGER` environment variable, to hydrate with data.

    Parameters
    ----------
    1. **symbol**: ``str`` 
        Symbol representing the statistic whose history is to be retrieved. List of allowable values can be found [here](https://www.quandl.com/data/FRED-Federal-Reserve-Economic-Data/documentation)
    2. **start_date**: ``Union[date, None]`` 
        *Optional*. Start date of price history. Defaults to None. If `start_date is None`, the calculation is made as if the `start_date` were set to 100 trading days ago. This excludes weekends and holidays.
    3. **end_date**: ``Union[date, None]``
        *Optional*. End date of price history. Defaults to None. If `end_date is None`, the calculation is made as if the `end_date` were set to today. This excludes weekends and holidays so that `end_date` is set to the last previous business date.

    Returns
    ------
    ``list``: `{ 'date' (str) :  value (str),  'date' (str):  value (str), ... }`
        Dictionary with date strings formatted `YYYY-MM-DD` as keys and the statistic on that date as the corresponding value.

    Raises
    ------
    1. **scrilla.errors.PriceError**
        If no sample prices can be retrieved, this error is thrown. 

    .. notes::
        * Most financial statistics are not reported on weekends or holidays, so the `asset_type` for financial statistics is functionally equivalent to equities, at least as far as date calculations are concerned. The dates inputted into this function are validated as if they were labelled as equity `asset_types` for this reason.

    """

    start_date, end_date = errors.validate_dates(
        start_date=start_date, end_date=end_date, asset_type=keys.keys['ASSETS']['EQUITY'])

    stats = stat_manager.get_stats(
        symbol=symbol, start_date=start_date, end_date=end_date)

    if not stats:
        raise errors.PriceError(
            f'Prices could not be retrieved for {symbol}')

    return stats


def get_daily_fred_latest(symbol: str) -> float:
    """
    Returns the latest value for the inputted statistic symbol.

    Parameters
    ----------
    1. **symbol**: str
        Symbol representing the statistic whose value it to be retrieved.
    """
    stats_history = get_daily_fred_history(symbol=symbol)
    first_element = helper.get_first_json_key(stats_history)
    return stats_history[first_element]


def get_daily_interest_history(maturity: str, start_date: Union[date, None] = None, end_date: Union[date, None] = None) -> list:
    """
    Wrapper around external service request for US Treasury Yield Curve data. Relies on an instance of `StatManager` configured by `settings.STAT_MANAGER` value, which in turn is configured by the `STAT_MANAGER` environment variable, to hydrate with data.

    Before deferring to the `StatManager` and letting it call the external service, however, this function checks if response is in local cache. If the response is not in the cache, it will pass the request off to `StatManager` and then save the response in the cache so subsequent calls to the function can bypass the service request. Used to prevent excessive external HTTP requests and improve the performance of the application. Other parts of the program should interface with the external statistics data services through this function to utilize the cache functionality. 

    Parameters
    ----------
    1. **maturity** : ``str``
        Maturity of the US Treasury for which the interest rate will be retrieved. List of allowable values can in `scrilla.stats.keys['SERVICES']['STATISTICS']['QUANDL']['MAP']['YIELD_CURVE']`
    2. **start_date** : ``datetime.date``
        *Optional*. Start date of price history. Defaults to None. If `start_date is None`, the calculation is made as if the `start_date` were set to 100 trading days ago. This excludes weekends and holidays.
    3. **end_date** : ``datetime.date``
        *Optional*. End date of price history. Defaults to None. If `end_date is None`, the calculation is made as if the `end_date` were set to today. This excludes weekends and holidays so that `end_date` is set to the last previous business date.

    Returns
    ------
    ``dict`` : `{ 'date' :  value ,  'date':  value , ... }`
        Dictionary with date strings formatted `YYYY-MM-DD` as keys and the interest on that date as the corresponding value.

    .. notes::
        * Yield rates are not reported on weekends or holidays, so the `asset_type` for interest is functionally equivalent to equities, at least as far as date calculations are concerned. The dates inputted into this function are validated as if they were labelled as equity `asset_types` for this reason.
    """
    start_date, end_date = errors.validate_dates(
        start_date=start_date, end_date=end_date, asset_type=keys.keys['ASSETS']['EQUITY'])

    rates = None
    rates = interest_cache.filter_interest_cache(
        maturity, start_date=start_date, end_date=end_date)

    if rates is not None:
        logger.debug(
            f'Comparing {len(rates)} = {dater.business_days_between(start_date, end_date)}')

    # TODO: this only works when stats are reported daily and that the latest date in the dataset is actually end_date.
    if rates is not None and \
            dater.to_string(end_date) in rates.keys() and \
            dater.business_days_between(start_date, end_date) == len(rates):
        return rates

    logger.debug(
        f'Cached {maturity} data is out of date, passing request to external service')
    rates = stat_manager.get_interest_rates(
        start_date=start_date, end_date=end_date)

    for this_date in rates:
        interest_cache.save_row(date=this_date, value=rates[this_date])

    rates = stat_manager.format_for_maturity(maturity=maturity, results=rates)

    return rates


def get_daily_interest_latest(maturity: str) -> float:
    """
    Returns the latest interest rate for the inputted US Treasury maturity.

    Parameters
    ----------
    1. **maturity**: ``str``
        Maturity of the US Treasury security whose interest rate is to be retrieved. Allowable values accessible through `keys.keys['YIELD_CURVE']
    """
    end_date = dater.get_last_trading_date()
    start_date = dater.decrement_date_by_business_days(end_date, 1)
    interest_history = get_daily_interest_history(
        maturity=maturity, start_date=start_date, end_date=end_date)
    first_element = helper.get_first_json_key(interest_history)
    return interest_history[first_element]


def get_dividend_history(ticker: str) -> dict:
    """
    Wrapper around external service request for dividend payment data. Relies on an instance of `DivManager` configured by `settings.DIV_MANAGER` value, which in turn is configured by the `DIV_MANAGER` environment variable, to hydrate with data.

    Note, since dividend payments do not occur every day (if only), dividend amounts do not get cached, as there is no nice way to determine on a given day whether or not a payment should have been made, and thus to determine whether or not the cache is out of date. In other words, you can't look at today's date and the date of the last payment in the cache and determine based solely on the dates whether or not the cache is outdated. 

    Parameters
    ----------
    1. **ticker** : ``str`` 
        Ticker symbol of the equity whose dividend history is to be retrieved.

    Returns
    ------
    ``list`` : `{ 'date' (str) :  amount (str),  'date' (str):  amount (str), ... }`
        Dictionary with date strings formatted `YYYY-MM-DD` as keys and the dividend payment amount on that date as the corresponding value.
    """
    logger.debug(f'Retrieving {ticker} dividends from service')
    divs = div_manager.get_dividends(ticker=ticker)
    return divs


def get_risk_free_rate() -> float:
    """
    Returns the risk free rate, defined as the annualized yield on a specific US Treasury duration, as a decimal. The US Treasury yield used as a proxy for the risk free rate is defined in the `settings.py` file and is configured through the RISK_FREE environment variable.
    """
    return get_daily_interest_latest(maturity=settings.RISK_FREE_RATE)/100



# This file is part of scrilla: https://github.com/chinchalinchin/scrilla.

# scrilla is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.

# scrilla is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with scrilla.  If not, see <https://www.gnu.org/licenses/>
# or <https://github.com/chinchalinchin/scrilla/blob/develop/main/LICENSE>.
"""
A module of functions that calculate financial statistics.
"""
import datetime
from typing import Dict, Union
from datetime import date
from scrilla import settings, services, files, cache, errors
from scrilla.static import keys
from scrilla.analysis.objects.cashflow import Cashflow
import scrilla.analysis.models.geometric.statistics as statistics
import scrilla.util.outputter as outputter

logger = outputter.Logger('scrilla.analysis.markets', settings.LOG_LEVEL)
profile_cache = cache.ProfileCache()


def sharpe_ratio(ticker: str, start_date: Union[date, None] = None, end_date: Union[date, None] = None, risk_free_rate: Union[float, None] = None, ticker_profile: Union[dict, None] = None, method: str = settings.ESTIMATION_METHOD) -> float:
    """
    Calculates the sharpe ratio for the supplied ticker over the specified time range. If no start and end date are supplied, calculation will default to the last 100 days of prices. The risk free rate and ticker risk profile can be passed in to force calculations without historical data.

    Parameters
    ----------
    1. **ticker**: ``str``
        A string of the ticker symbol whose sharpe ratio will be computed.
    2. **start_date**: ``datetime.date``
        Start date of the time period for which the sharpe ratio will be computed.
    3. **end_date**: ``datetime.date``
        End_date of the time period for which the sharpe ratio will be computed.
    4. **risk_free_rate**: ``float``
        Risk free rate used to evaluate excess return. Defaults to settings.RISK_FREE_RATE.
    5. **ticker_profile**: ``dict`` 
        Risk-return profile for the supplied ticker. Formatted as follows: `{ 'annual_return': float, 'annual_volatility': float } `
    6. **method** : ``str``
        Estimation method used to calculate financial statistics. Defaults to the value set by `scrilla.settings.ESTIMATION_METHOD`. Allowable value are accessible through the `scrilla.keys.keys` dictionary.

    .. notes:: 
        * if ``ticker_profile`` is provided, this function will skip both an external data service call and the calculation of the ticker's risk profile. The calculation will proceed as if the supplied profile were the true profile. If ``ticker_profile`` is not provided, all statistics will be estimated from historical data.
    """
    start_date, end_date = errors.validate_dates(start_date=start_date, end_date=end_date,
                                                 asset_type=keys.keys['ASSETS']['EQUITY'])

    if ticker_profile is None:
        result = profile_cache.filter_profile_cache(
            ticker=ticker, start_date=start_date, end_date=end_date, method=method)

        if result is not None and result[keys.keys['STATISTICS']['SHARPE']] is not None:
            return result[keys.keys['STATISTICS']['SHARPE']]

        ticker_profile = statistics.calculate_risk_return(
            ticker=ticker, start_date=start_date, end_date=end_date, method=method)

    if risk_free_rate is None:
        risk_free_rate = services.get_risk_free_rate()

    sh_ratio = (ticker_profile['annual_return'] -
                risk_free_rate)/ticker_profile['annual_volatility']

    profile_cache.save_or_update_row(ticker=ticker, start_date=start_date,
                                     end_date=end_date, sharpe_ratio=sh_ratio, method=method)

    return sh_ratio

# if no dates are specified, defaults to last 100 days


def market_premium(start_date: Union[date, None] = None, end_date: Union[date, None] = None, market_profile: Union[dict, None] = None, method: str = settings.ESTIMATION_METHOD) -> float:
    """
    Returns the excess of the market return defined by the environment variable `MARKET_PROXY` over the risk free rate defined by the `RISK_FREE` environment variable.

    Parameters
    ----------
    1. **start_date**: ``datetime.date``
        Start date of the time period for which the market premium will be computed.
    2. **end_date**: ``datetime.date``
        End_date of the time period for which the market premium will be computed.
    3. **market_profile**: ``dict``
    4. **method** : ``str``
        Estimation method used to calculate financial statistics. Defaults to the value set by `scrilla.settings.ESTIMATION_METHOD`. Allowable value are accessible through the `scrilla.keys.keys` dictionary.

    """
    start_date, end_date = errors.validate_dates(start_date=start_date, end_date=end_date,
                                                 asset_type=keys.keys['ASSETS']['EQUITY'])

    if market_profile is None:
        market_profile = statistics.calculate_risk_return(
            ticker=settings.MARKET_PROXY, start_date=start_date, end_date=end_date, method=method)

    market_prem = (
        market_profile['annual_return'] - services.get_risk_free_rate())
    return market_prem


def market_beta(ticker: str, start_date: Union[date, None] = None, end_date: Union[date, None] = None, market_profile: Union[dict, None] = None, market_correlation: Union[dict, None] = None, ticker_profile: Union[dict, None] = None, sample_prices: Union[dict, None] = None, method: str = settings.ESTIMATION_METHOD) -> float:
    """
    Returns the beta of an asset against the market return defined by the ticker symbol set `scrilla.settings.MARKET_PROXY`, which in turn is configured through the environment variable of the same name, `MARKET_PROXY`. 

    Parameters
    ----------
    1. **ticker**: ``str``
        A string of the ticker symbol whose asset beta will be computed.
    2. **start_date**: ``datetime.date``
        Start date of the time period for which the asset beta will be computed.
    3. **end_date**: ``datetime.date`` 
        End_date of the time period for which the asset beta will be computed.
    4. **method** : ``str``
        Estimation method used to calculate financial statistics. Defaults to the value set by `scrilla.settings.ESTIMATION_METHOD`. Allowable value are accessible through the `scrilla.keys.keys` dictionary.

    .. notes::
        * If not configured by an environment variable, `scrilla.settings.MARKET_PROXY` defaults to ``SPY``, the ETF tracking the *S&P500*.
    """
    start_date, end_date = errors.validate_dates(start_date=start_date, end_date=end_date,
                                                 asset_type=keys.keys['ASSETS']['EQUITY'])

    result = profile_cache.filter_profile_cache(
        ticker=ticker, start_date=start_date, end_date=end_date, method=method)

    if result is not None and result[keys.keys['STATISTICS']['BETA']] is not None:
        return result[keys.keys['STATISTICS']['BETA']]

    if market_profile is None:
        if sample_prices is None:
            market_profile = statistics.calculate_risk_return(ticker=settings.MARKET_PROXY, start_date=start_date,
                                                              end_date=end_date, method=method)
        else:
            market_profile = statistics.calculate_risk_return(ticker=settings.MARKET_PROXY, method=method,
                                                              sample_prices=sample_prices[settings.MARKET_PROXY])
    if ticker_profile is None:
        if sample_prices is None:
            ticker_profile = statistics.calculate_risk_return(ticker=ticker, start_date=start_date,
                                                              end_date=end_date, method=method)
        else:
            ticker_profile = statistics.calculate_risk_return(ticker=ticker, method=method,
                                                              sample_prices=sample_prices[ticker])

    market_covariance = statistics.calculate_return_covariance(ticker_1=ticker, ticker_2=settings.MARKET_PROXY,
                                                               profile_1=ticker_profile, profile_2=market_profile,
                                                               correlation=market_correlation,
                                                               sample_prices=sample_prices,
                                                               start_date=start_date, end_date=end_date)

    beta = market_covariance / (market_profile['annual_volatility']**2)

    profile_cache.save_or_update_row(
        ticker=ticker, start_date=start_date, end_date=end_date, asset_beta=beta, method=method)

    return beta


def cost_of_equity(ticker: str, start_date: Union[datetime.date, None] = None, end_date: Union[datetime.date, None] = None, market_profile: Union[Dict[str, float], None] = None, market_correlation: Union[Dict[str, float], None] = None, method=settings.ESTIMATION_METHOD) -> float:
    """
    Returns the cost of equity of an asset as estimated by the Capital Asset Pricing Model, i.e. the product of the market premium and asset beta increased by the risk free rate.

    Parameters
    ----------
    1. **ticker**: ``str``
        A string of the ticker symbol whose cost of equity ratio will be computed.
    2. **start_date**: ``Union[datetime.date, None]``
        *Optional*. Start date of the time period for which the cost of equity ratio will be computed
    3. **end_date**: ``Union[datetime.date, None]``
        *Optional.* End_date of the time period for which the cost of equity ratio will be computed.
    4. **market_profile**: ``Union[Dict[str, float], None]``
        *Optional*. Dictionary containing the assumed risk profile for the market proxy. Overrides calls to services and staistical methods, forcing the calculation fo the cost of equity with the inputted market profile. Format: ``{ 'annual_return': value, 'annual_volatility': value}``
    5. **market_correlation**: ``Union[Dict[str, float], None]``
        *Optional*. Dictionary containing the assumed correlation for the calculation. Overrides calls to services and statistical methods, forcing the calculation of the cost of equity with the inputted correlation. Format: ``{ 'correlation' : value }``
    6. **method** : ``str``
        *Optional*. Estimation method used to calculate financial statistics. Defaults to the value set by `scrilla.settings.ESTIMATION_METHOD`. Allowable value are accessible through the `scrilla.keys.keys` dictionary.
    """
    start_date, end_date = errors.validate_dates(start_date=start_date, end_date=end_date,
                                                 asset_type=keys.keys['ASSETS']['EQUITY'])

    result = profile_cache.filter_profile_cache(
        ticker=ticker, start_date=start_date, end_date=end_date, method=method)

    if result is not None and result[keys.keys['STATISTICS']['EQUITY']] is not None:
        return result[keys.keys['STATISTICS']['EQUITY']]

    beta = market_beta(ticker=ticker, start_date=start_date, end_date=end_date,
                       market_profile=market_profile, market_correlation=market_correlation,
                       method=method)
    premium = market_premium(start_date=start_date, end_date=end_date,
                             market_profile=market_profile, method=method)

    equity_cost = (premium*beta + services.get_risk_free_rate())

    profile_cache.save_or_update_row(
        ticker=ticker, start_date=start_date, end_date=end_date, equity_cost=equity_cost, method=method)
    return equity_cost


def screen_for_discount(model: str = keys.keys['MODELS']['DDM'], discount_rate: float = None) -> Dict[str, Dict[str, float]]:
    """
    Screens the stocks saved under the user watchlist in the `scrilla.settings.COMMON_DIR` directory for discounts relative to the model inputted into the function.

    Parameters
    ----------
    1. **model** : ``str``
        *Optional*. Model used to evaluated the equities saved in the watchlist. If no model is specified, the function will default to the discount dividend model. Model constants are accessible through the the `scrilla.keys.keys` dictionary.
    2. **discount_rate** : ``float``
        *Optional*. Rate used to discount future cashflows to present. Defaults to an equity's CAPM cost of equity, as calculated by `scrilla.analysis.markets.cost_of_equity`.

    Returns
    -------
    ``dict``
        A list of tickers that trade at a discount relative to the model price, formatted as follows: `{ 'ticker' : { 'spot_price': value, 'model_price': value,'discount': value }, ... }`
    """

    equities = list(files.get_watchlist())
    discounts = {}
    user_discount_rate = discount_rate

    for equity in equities:
        spot_price = services.get_daily_price_latest(ticker=equity)

        if user_discount_rate is None:
            discount_rate = cost_of_equity(ticker=equity)
        else:
            discount_rate = user_discount_rate

        if model == keys.keys['MODELS']['DDM']:
            logger.debug(
                'Using Discount Dividend Model to screen watchlisted equities for discounts.')
            dividends = services.get_dividend_history(equity)
            model_price = Cashflow(
                sample=dividends, discount_rate=discount_rate).calculate_net_present_value()
            discount = float(model_price) - float(spot_price)

            if discount > 0:
                discount_result = {}
                discount_result['spot_price'] = spot_price
                discount_result['model_price'] = model_price
                discount_result['discount'] = discount
                discounts[equity] = discount_result
                logger.debug(f'Discount of {discount} found for {equity}')

    return discounts



# This file is part of scrilla: https://github.com/chinchalinchin/scrilla.

# scrilla is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.

# scrilla is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with scrilla.  If not, see <https://www.gnu.org/licenses/>
# or <https://github.com/chinchalinchin/scrilla/blob/develop/main/LICENSE>.

from typing import Union
from PySide6 import QtGui, QtCore, QtWidgets


from scrilla import settings, services
from scrilla.static import keys, definitions
# TODO: conditional import based on ANALYSIS_MODE
from scrilla.analysis import estimators, markets, optimizer, plotter
from scrilla.analysis.models.geometric import statistics
from scrilla.analysis.objects.portfolio import Portfolio
from scrilla.analysis.objects.cashflow import Cashflow

from scrilla.util import dater, outputter, helper

from scrilla.gui import formats, utilities
from scrilla.gui.widgets import factories, components

logger = outputter.Logger('gui.functions', settings.LOG_LEVEL)


class DistributionWidget(components.SkeletonWidget):
    def __init__(self, layer: str, parent: Union[QtWidgets.QWidget, None] = None):
        super().__init__(function='plot_return_dist', parent=parent)
        self.setObjectName(layer)
        self._init_widgets()
        self._arrange_widgets()
        self._stage_widgets()

    def _init_widgets(self):
        self.arg_widget = components.ArgumentWidget(calculate_function=self.calculate,
                                                    clear_function=self.clear,
                                                    controls=self.controls,
                                                    layer=utilities.get_next_layer(self.objectName()))
        self.title = factories.atomic_widget_factory(
            component='heading', title='Distribution of Returns')
        self.tab_container = factories.layout_factory(layout='vertical-box')
        self.tab_widget = QtWidgets.QTabWidget()
        self.setLayout(QtWidgets.QHBoxLayout())

    def _arrange_widgets(self):
        self.tab_container.layout().addWidget(self.tab_widget)
        self.layout().addWidget(self.tab_container)
        self.layout().addWidget(self.arg_widget)

    def _stage_widgets(self):
        self.arg_widget.prime()

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        for i in range(self.tab_widget.count()):
            if self.tab_widget.widget(i).figure.isVisible():
                self.tab_widget.widget(i).set_pixmap()
        return super().resizeEvent(event)

    @QtCore.Slot()
    def calculate(self):
        symbols = self.arg_widget.get_symbol_input()
        start_date = self.arg_widget.get_control_input('start_date')
        end_date = self.arg_widget.get_control_input('end_date')

        for symbol in symbols:
            qq_plot = components.GraphWidget(tmp_graph_key=f'{keys.keys["GUI"]["TEMP"]["QQ"]}_{symbol}',
                                             layer=utilities.get_next_layer(self.objectName()))
            dist_plot = components.GraphWidget(tmp_graph_key=f'{keys.keys["GUI"]["TEMP"]["DIST"]}_{symbol}',
                                               layer=utilities.get_next_layer(self.objectName()))
            returns = statistics.get_sample_of_returns(ticker=symbol,
                                                       start_date=start_date,
                                                       end_date=end_date,
                                                       daily=True)
            qq_series = estimators.qq_series_for_sample(sample=returns)
            plotter.plot_qq_series(ticker=symbol,
                                   qq_series=qq_series,
                                   show=False,
                                   savefile=f'{settings.TEMP_DIR}/{keys.keys["GUI"]["TEMP"]["QQ"]}_{symbol}')
            plotter.plot_return_histogram(ticker=symbol,
                                          sample=returns,
                                          show=False,
                                          savefile=f'{settings.TEMP_DIR}/{keys.keys["GUI"]["TEMP"]["DIST"]}_{symbol}',)
            dist_plot.set_pixmap()
            qq_plot.set_pixmap()
            self.tab_widget.addTab(qq_plot, f'{symbol} QQ Plot')
            self.tab_widget.addTab(dist_plot, f'{symbol} Distribution')
        self.tab_widget.show()
        self.arg_widget.fire()

    @QtCore.Slot()
    def clear(self):
        self.arg_widget.prime()
        total = self.tab_widget.count()
        for _ in range(total):
            self.tab_widget.removeTab(0)


class YieldCurveWidget(components.SkeletonWidget):
    def __init__(self, layer: str, parent: Union[QtWidgets.QWidget, None] = None):
        super().__init__(function='yield_curve', parent=parent)
        self.setObjectName(layer)
        self._init_widgets()
        self._arrange_widgets()
        self._stage_widgets()

    def _init_widgets(self):
        self.graph_widget = components.GraphWidget(tmp_graph_key=keys.keys['GUI']['TEMP']['YIELD'],
                                                   layer=utilities.get_next_layer(self.objectName()))
        self.arg_widget = components.ArgumentWidget(calculate_function=self.calculate,
                                                    clear_function=self.clear,
                                                    controls=self.controls,
                                                    layer=utilities.get_next_layer(
                                                        self.objectName()),
                                                    mode=components.SYMBOLS_NONE)
        # TODO: initialize arg widget WITHOUT tickers

        self.setLayout(QtWidgets.QHBoxLayout())

    def _arrange_widgets(self):
        self.layout().addWidget(self.graph_widget)
        self.layout().addWidget(self.arg_widget)

    def _stage_widgets(self):
        self.arg_widget.prime()

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        if self.graph_widget.figure.isVisible():
            self.graph_widget.set_pixmap()
        return super().resizeEvent(event)

    @QtCore.Slot()
    def calculate(self):
        if self.graph_widget.figure.isVisible():
            self.graph_widget.figure.hide()

        yield_curve = {}
        start_date = dater.this_date_or_last_trading_date(
            self.arg_widget.get_control_input('start_date'))
        start_string = dater.to_string(start_date)
        yield_curve[start_string] = []
        for maturity in keys.keys['YIELD_CURVE']:
            rate = services.get_daily_interest_history(maturity=maturity,
                                                       start_date=start_date,
                                                       end_date=start_date)
            yield_curve[start_string].append(rate[start_string])

        plotter.plot_yield_curve(yield_curve=yield_curve,
                                 show=False,
                                 savefile=f'{settings.TEMP_DIR}/{keys.keys["GUI"]["TEMP"]["YIELD"]}')
        self.graph_widget.set_pixmap()
        self.arg_widget.fire()

    @QtCore.Slot()
    def clear(self):
        self.graph_widget.clear()
        self.arg_widget.prime()


class DiscountDividendWidget(components.SkeletonWidget):
    def __init__(self, layer: str, parent: Union[QtWidgets.QWidget, None] = None):
        super().__init__(function='discount_dividend', parent=parent)
        self.setObjectName(layer)
        self._init_widgets()
        self._arrange_widgets()
        self._stage_widgets()

    def _init_widgets(self):
        self.tab_container = factories.layout_factory(layout='vertical-box')
        self.tab_widget = QtWidgets.QTabWidget()
        self.arg_widget = components.ArgumentWidget(calculate_function=self.calculate,
                                                    clear_function=self.clear,
                                                    controls=self.controls,
                                                    layer=utilities.get_next_layer(self.objectName()))
        # TODO: restrict arg symbol input to one symbol somehow
        self.setLayout(QtWidgets.QHBoxLayout())

    def _arrange_widgets(self):
        self.tab_container.layout().addWidget(self.tab_widget)
        self.layout().addWidget(self.tab_container)
        self.layout().addWidget(self.arg_widget)

    def _stage_widgets(self):
        self.arg_widget.prime()

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        for i in range(self.tab_widget.count()):
            if self.tab_widget.widget(i).figure.isVisible():
                self.tab_widget.widget(i).set_pixmap()
        return super().resizeEvent(event)

    @QtCore.Slot()
    def calculate(self):
        symbols = self.arg_widget.get_symbol_input()
        discount = self.arg_widget.get_control_input('discount')

        for symbol in symbols:
            if discount is None:
                discount = markets.cost_of_equity(ticker=symbol,
                                                  start_date=self.arg_widget.get_control_input(
                                                      'start_date'),
                                                  end_date=self.arg_widget.get_control_input('end_date'))
            dividends = services.get_dividend_history(ticker=symbol)
            cashflow = Cashflow(sample=dividends, discount_rate=discount)
            graph_widget = components.GraphWidget(tmp_graph_key=f'{keys.keys["GUI"]["TEMP"]["DIVIDEND"]}_{symbol}',
                                                  layer=utilities.get_next_layer(self.objectName()))
            plotter.plot_cashflow(ticker=symbol,
                                  cashflow=cashflow,
                                  show=False,
                                  savefile=f'{settings.TEMP_DIR}/{keys.keys["GUI"]["TEMP"]["DIVIDEND"]}_{symbol}')

            graph_widget.set_pixmap()
            self.tab_widget.addTab(graph_widget, f'{symbol} DDM PLOT')
        self.tab_widget.show()
        self.arg_widget.fire()

    @QtCore.Slot()
    def clear(self):
        self.arg_widget.prime()
        total = self.tab_widget.count()
        for _ in range(total):
            self.tab_widget.removeTab(0)


class RiskProfileWidget(components.SkeletonWidget):
    def __init__(self, layer: str, parent: Union[QtWidgets.QWidget, None] = None):
        super().__init__(function='risk_profile', parent=parent)
        self.setObjectName(layer)
        self._init_widgets()
        self._arrange_widgets()
        self._stage_widgets()

    def _init_widgets(self):
        self.composite_widget = components.CompositeWidget(keys.keys['GUI']['TEMP']['PROFILE'],
                                                           widget_title="Risk Analysis",
                                                           table_title="CAPM Risk Profile",
                                                           graph_title="Risk-Return Plane",
                                                           layer=utilities.get_next_layer(self.objectName()))
        self.arg_widget = components.ArgumentWidget(calculate_function=self.calculate,
                                                    clear_function=self.clear,
                                                    controls=self.controls,
                                                    layer=utilities.get_next_layer(self.objectName()))
        self.setLayout(QtWidgets.QHBoxLayout())

    def _arrange_widgets(self):
        self.layout().addWidget(self.composite_widget)
        self.layout().addWidget(self.arg_widget)

    def _stage_widgets(self):
        self.arg_widget.prime()

    @QtCore.Slot()
    def calculate(self):
        if self.composite_widget.graph_widget.figure.isVisible():
            self.composite_widget.graph_widget.figure.hide()

        symbols = self.arg_widget.get_symbol_input()

        self.composite_widget.table_widget.init_table(rows=symbols,
                                                      columns=['Return', 'Volatility', 'Sharpe', 'Beta', 'Equity Cost'])

        profiles = {}
        for i, symbol in enumerate(symbols):
            profiles[symbol] = statistics.calculate_risk_return(ticker=symbol,
                                                                start_date=self.arg_widget.get_control_input(
                                                                    'start_date'),
                                                                end_date=self.arg_widget.get_control_input('end_date'))
            profiles[symbol][keys.keys['APP']['PROFILE']
                             ['SHARPE']] = markets.sharpe_ratio(symbol)
            profiles[symbol][keys.keys['APP']['PROFILE']
                             ['BETA']] = markets.market_beta(symbol)
            profiles[symbol][keys.keys['APP']['PROFILE']
                             ['EQUITY']] = markets.cost_of_equity(symbol)

            formatted_profile = formats.format_profile(profiles[symbol])

            for j, statistic in enumerate(formatted_profile.keys()):
                table_item = factories.atomic_widget_factory(
                    component='table-item', title=formatted_profile[statistic])
                self.composite_widget.table_widget.table.setItem(
                    i, j, table_item)

        plotter.plot_profiles(symbols=symbols, profiles=profiles, show=False,
                              savefile=f'{settings.TEMP_DIR}/{keys.keys["GUI"]["TEMP"]["PROFILE"]}')

        self.composite_widget.graph_widget.set_pixmap()
        self.composite_widget.table_widget.show_table()
        self.arg_widget.fire()

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        if self.composite_widget.graph_widget.figure.isVisible():
            self.composite_widget.graph_widget.set_pixmap()
        return super().resizeEvent(event)

    @QtCore.Slot()
    def clear(self):
        self.composite_widget.graph_widget.clear()
        self.composite_widget.table_widget.table.clear()
        self.composite_widget.table_widget.table.hide()
        self.arg_widget.prime()


class CorrelationWidget(components.SkeletonWidget):
    def __init__(self, layer: str, parent: Union[QtWidgets.QWidget, None] = None):
        super().__init__(function='correlation', parent=parent)
        self.setObjectName(layer)
        self._init_widgets()
        self._arrange_widgets()
        self._stage_widgets()

    def _init_widgets(self):
        self.table_widget = components.TableWidget(widget_title="Correlation Matrix",
                                                   layer=utilities.get_next_layer(self.objectName()))
        self.arg_widget = components.ArgumentWidget(calculate_function=self.calculate,
                                                    clear_function=self.clear,
                                                    controls=self.controls,
                                                    layer=utilities.get_next_layer(self.objectName()))
        self.setLayout(QtWidgets.QHBoxLayout())

    def _arrange_widgets(self):
        self.layout().addWidget(self.table_widget)
        self.layout().addWidget(self.arg_widget)

    def _stage_widgets(self):
        self.arg_widget.prime()

    @QtCore.Slot()
    def calculate(self):
        if self.table_widget.table.isVisible():
            self.table_widget.table.clear()
            self.table_widget.table.hide()

        symbols = self.arg_widget.get_symbol_input()

        if len(symbols) > 1:
            self.table_widget.init_table(rows=symbols, columns=symbols)

            matrix = statistics.correlation_matrix(tickers=symbols,
                                                   start_date=self.arg_widget.get_control_input(
                                                       'start_date'),
                                                   end_date=self.arg_widget.get_control_input('end_date'))
            for i in range(0, len(symbols)):
                for j in range(i, len(symbols)):
                    item_upper = factories.atomic_widget_factory(
                        component='table-item', title=helper.format_float_percent(matrix[i][j]))
                    item_lower = factories.atomic_widget_factory(
                        component='table-item', title=helper.format_float_percent(matrix[j][i]))
                    self.table_widget.table.setItem(j, i, item_upper)
                    self.table_widget.table.setItem(i, j, item_lower)
        else:
            print('error handling goes here')

        self.table_widget.show_table()
        self.arg_widget.fire()

    @QtCore.Slot()
    def clear(self):
        self.arg_widget.prime()
        self.table_widget.table.clear()
        self.table_widget.table.hide()
        self.table_widget.download_button.hide()


class OptimizerWidget(components.SkeletonWidget):
    def __init__(self, layer: str, parent: Union[QtWidgets.QWidget, None] = None):
        super().__init__(function='optimize_portfolio', parent=parent)
        self.setObjectName(layer)
        self._init_widgets()
        self._arrange_widgets()
        self._stage_widgets()

    def _init_widgets(self):
        self.title = factories.atomic_widget_factory(
            component='heading', title=None)
        self.table_widget = components.TableWidget(widget_title="Optimization Results",
                                                   layer=utilities.get_next_layer(self.objectName()))
        self.arg_widget = components.ArgumentWidget(calculate_function=self.optimize,
                                                    clear_function=self.clear,
                                                    controls=self.controls,
                                                    layer=utilities.get_next_layer(self.objectName()))
        self.setLayout(QtWidgets.QHBoxLayout())

    def _arrange_widgets(self):
        self.layout().addWidget(self.table_widget)
        self.layout().addWidget(self.arg_widget)

    def _stage_widgets(self):
        self.arg_widget.prime()

    @QtCore.Slot()
    def optimize(self):
        if self.table_widget.table.isVisible():
            self.table_widget.table.clear()
            self.table_widget.table.hide()

        symbols = self.arg_widget.get_symbol_input()

        # TODO: better error checking
        if len(symbols) > 1:
            investment = self.arg_widget.get_control_input('investment')
            this_portfolio = Portfolio(tickers=symbols,
                                       start_date=self.arg_widget.get_control_input(
                                           'start_date'),
                                       end_date=self.arg_widget.get_control_input('end_date'))
            allocation = optimizer.optimize_portfolio_variance(portfolio=this_portfolio,
                                                               target_return=self.arg_widget.get_control_input('target'))
            self.title.setText(formats.format_allocation_profile_title(
                allocation, this_portfolio))

            prices = services.get_daily_prices_latest(tickers=symbols)

            if investment is None:
                self.table_widget.init_table(
                    rows=symbols, columns=['Allocation'])
            else:
                self.table_widget.init_table(rows=symbols, columns=[
                                             'Allocation', 'Shares'])
                shares = this_portfolio.calculate_approximate_shares(
                    allocation, float(investment), prices)

            for i in range(len(symbols)):
                item = factories.atomic_widget_factory(
                    component='table-item', title=helper.format_float_percent(allocation[i]))
                self.table_widget.table.setItem(i, 0, item)

                if investment is not None:
                    share_item = factories.atomic_widget_factory(
                        component='table-item', title=str(shares[i]))
                    self.table_widget.table.setItem(i, 1, share_item)

            # TODO: display amount vested per equity
            # TODO: display total portfolio return and volatility
            # TODO: display actual investment
            self.table_widget.show_table()
            self.arg_widget.fire()

        else:
            print('something went wrong')

    @QtCore.Slot()
    def clear(self):
        self.table_widget.table.clear()
        self.table_widget.table.hide()
        self.arg_widget.prime()


class EfficientFrontierWidget(components.SkeletonWidget):
    def __init__(self, layer: str, parent: Union[QtWidgets.QWidget, None] = None):
        super().__init__(function='efficient_frontier', parent=parent)
        self.setObjectName(layer)
        self._init_widgets()
        self._arrange_widgets()
        self._stage_widgets()

    def _init_widgets(self):
        self.graph_widget = components.GraphWidget(tmp_graph_key=keys.keys['GUI']['TEMP']['FRONTIER'],
                                                   layer=utilities.get_next_layer(self.objectName()))
        self.arg_widget = components.ArgumentWidget(calculate_function=self.calculate,
                                                    clear_function=self.clear,
                                                    controls=self.controls,
                                                    layer=utilities.get_next_layer(self.objectName()))
        self.setLayout(QtWidgets.QHBoxLayout())
        # TODO: portfolio tabs

    def _arrange_widgets(self):
        self.layout().addWidget(self.graph_widget)
        self.layout().addWidget(self.arg_widget)

    def _stage_widgets(self):
        self.arg_widget.prime()

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        if self.graph_widget.figure.isVisible():
            self.graph_widget.set_pixmap()
        return super().resizeEvent(event)

    @QtCore.Slot()
    def calculate(self):
        if self.graph_widget.figure.isVisible():
            self.graph_widget.figure.hide()

        this_portfolio = Portfolio(tickers=self.arg_widget.get_symbol_input(),
                                   start_date=self.arg_widget.get_control_input(
                                       'start_date'),
                                   end_date=self.arg_widget.get_control_input('end_date'))
        frontier = optimizer.calculate_efficient_frontier(portfolio=this_portfolio,
                                                          steps=self.arg_widget.get_control_input('steps'))
        plotter.plot_frontier(portfolio=this_portfolio,
                              frontier=frontier,
                              show=False,
                              savefile=f'{settings.TEMP_DIR}/{keys.keys["GUI"]["TEMP"]["FRONTIER"]}')

        self.graph_widget.set_pixmap()
        self.arg_widget.fire()

    @QtCore.Slot()
    def clear(self):
        self.graph_widget.figure.hide()
        self.arg_widget.prime()


class MovingAverageWidget(components.SkeletonWidget):
    def __init__(self, layer: str, parent: Union[QtWidgets.QWidget, None] = None):
        super().__init__(function='moving_averages', parent=parent)
        self.setObjectName(layer)
        self._init_widgets()
        self._arrange_widgets()
        self._stage_widgets()

    def _init_widgets(self):
        self.graph_widget = components.GraphWidget(keys.keys['GUI']['TEMP']['AVERAGES'],
                                                   layer=utilities.get_next_layer(self.objectName()))
        self.arg_widget = components.ArgumentWidget(calculate_function=self.calculate,
                                                    clear_function=self.clear,
                                                    controls=self.controls,
                                                    layer=utilities.get_next_layer(
                                                        self.objectName()),
                                                    mode=components.SYMBOLS_SINGLE)
        self.setLayout(QtWidgets.QHBoxLayout())

    def _arrange_widgets(self):
        self.layout().addWidget(self.graph_widget)
        self.layout().addWidget(self.arg_widget)

    def _stage_widgets(self):
        self.arg_widget.prime()

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        if self.graph_widget.figure.isVisible():
            self.graph_widget.set_pixmap()
        return super().resizeEvent(event)

    @QtCore.Slot()
    def calculate(self):
        if self.graph_widget.figure.isVisible():
            self.graph_widget.figure.hide()

        moving_averages = statistics.calculate_moving_averages(ticker=self.arg_widget.get_symbol_input()[0],
                                                               start_date=self.arg_widget.get_control_input(
                                                                   'start_date'),
                                                               end_date=self.arg_widget.get_control_input('end_date'))

        plotter.plot_moving_averages(ticker=self.arg_widget.get_symbol_input()[0],
                                     averages=moving_averages,
                                     show=False,
                                     savefile=f'{settings.TEMP_DIR}/{keys.keys["GUI"]["TEMP"]["AVERAGES"]}')
        self.graph_widget.set_pixmap()
        self.arg_widget.fire()

    @QtCore.Slot()
    def clear(self):
        self.arg_widget.prime()
        self.graph_widget.clear()



from scrilla.static.constants import constants

formats = {
    'separator': '-',
    'TAB': '     ',
    'LINE_LENGTH': 100,
    'BAR_WIDTH': 0.10,
    'INDENT': 10,
    'RISK_FREE_TITLE': "{} US Treasury",
    'BINS': 20
}


def format_allocation(allocation, portfolio, investment=None, latest_prices=None):
    allocation_format = []

    if investment is not None:
        shares = portfolio.calculate_approximate_shares(
            x=allocation, total=investment, latest_prices=latest_prices)
        total = portfolio.calculate_actual_total(
            x=allocation, total=investment, latest_prices=latest_prices)

    annual_volatility = portfolio.volatility_function(x=allocation)
    annual_return = portfolio.return_function(x=allocation)

    for j, item in enumerate(portfolio.tickers):
        holding = {}
        holding['ticker'] = item
        holding['allocation'] = round(allocation[j], constants['ACCURACY'])
        if investment is not None:
            holding['shares'] = float(shares[j])
        holding['annual_return'] = round(
            portfolio.mean_return[j], constants['ACCURACY'])
        holding['annual_volatility'] = round(
            portfolio.sample_vol[j], constants['ACCURACY'])
        allocation_format.append(holding)

    json_format = {}
    json_format['holdings'] = allocation_format

    if investment is not None:
        json_format['total'] = float(total)

    json_format['portfolio_return'] = annual_return
    json_format['portfolio_volatility'] = annual_volatility

    return json_format


def format_frontier(portfolio, frontier, investment=None, latest_prices=None):
    json_format = []
    for item in frontier:
        json_format.append(format_allocation(allocation=item, portfolio=portfolio,
                                             investment=investment, latest_prices=latest_prices))
    return json_format


def format_correlation_matrix(tickers, correlation_matrix):
    response = []
    for i, item in enumerate(tickers):
        # correlation_matrix[i][i]
        for j in range(i+1, len(tickers)):
            subresponse = {}
            subresponse[f'{item}_{tickers[j]}_correlation'] = correlation_matrix[j][i]
            response.append(subresponse)
    return response



import datetime
from typing import Dict, List, Tuple, Union

from scrilla.static import constants, formats, definitions


# GENERAL OUTPUT FUNCTIONS
def space(n: int):
    print('\n'*n)


def title_line(title: str, line_length: int = formats.formats['LINE_LENGTH'], separator: str = formats.formats['separator'],
               display: bool = True):
    buff = int((line_length - len(title))/2)
    result = separator*buff + title + separator*buff
    if display:
        print(result)
    return result


def separator_line(line_length=formats.formats['LINE_LENGTH'], separator=formats.formats['separator'], display=True) -> str:
    if display:
        print(separator*line_length)
    return separator*line_length


def break_lines(msg: str, line_length: int = formats.formats['LINE_LENGTH']) -> List[str]:
    """
    Generates a list of strings where each string is less than or equal to a specified line length. All elements in the array will be equal length except the last element; in other words, the inputted string will be read into an array, line by line, until there is nothing left to be read.

    Parameters
    ----------
    1. ``msg``: ``str``
        String to be broken into an array with each string no longer than `line_length`.
    2. **line_length**: ``int``
        *Optional*. Defaults to the value set by `static.formats.formats['LINE_LENGTH]`. The maximum length of a line. 

    Returns
    -------
    `list[str]`
        A list of `line`'s with `len(line)<line_length or len(line)==line_length`.
    """

    if len(msg) > line_length:
        return [msg[i:i+line_length] for i in range(0, len(msg), line_length)]
    return [msg]


def center(this_line: str, line_length: int = formats.formats['LINE_LENGTH'], display: bool = True) -> str:
    """
    Centers an inputted line relative to the line length. 

    Parameters
    ----------
    1. **this_line**: ``str``
        the string to centered.
    2. **line_length**: ``int``
        *Optional*. Defaults to the value set by `static.formats.formats['LINE_LENGTH]`. The length of the line that is being centered over. 
    3. **display**: ``bool``
        *Optional*. Defaults to `True`. If set to `True`, function will print result to screen. If set to `False`, the function will return the value as a string.

    Returns
    -------
    `str`
        The line with the difference between `line_length` and `this_line` distributed evenly on either side of the content as white space. 

    Raises
    ------
    1. **ValueError**
        If the inputted line is larged than the line length, this error will be thrown.
    """

    if len(this_line) > line_length:
        raise ValueError(
            'Line to be centered is larger than length being centered over.')

    buff = int((line_length - len(this_line))/2)
    output = ' '*buff + this_line + ' '*buff
    if display:
        print(output)
        return
    return output


def print_list(list_to_print, indent=formats.formats['INDENT']):
    for i, item in enumerate(list_to_print):
        print(' '*indent, f'{i}. {item}')


def string_result(operation, result, indent=formats.formats['INDENT'], display=True):
    output = ' '*indent + operation + ' = ' + result
    if display:
        print(output)
        return
    return output


def scalar_result(calculation, result, currency=True, indent=formats.formats['INDENT']):
    if currency:
        print(' '*indent, calculation, ' = $', round(float(result), 2))
    else:
        print(' '*indent, calculation, ' = ', round(float(result), 4))


def percent_result(calculation, result, indent=formats.formats['INDENT']):
    print(' '*indent, calculation, ' = ', round(float(result), 4), '%')


def equivalent_result(right_hand, left_hand, value, indent=formats.formats['INDENT']):
    print(' '*indent, f'{right_hand} = {left_hand} = {value}')


def help_msg(indent=formats.formats['INDENT']):
    func_dict, arg_dict = definitions.FUNC_DICT, definitions.ARG_DICT

    title_line('scrilla')
    space(1)

    for paragraph in definitions.HELP_MSG:
        for line in break_lines(paragraph):
            center(line)
        space(1)

    title_line('SYNTAX')
    center(definitions.SYNTAX)
    space(1)

    for func_name in func_dict:
        title_line(func_dict[func_name]['name'])
        for line in break_lines(func_dict[func_name]['description']):
            center(line)
        separator_line()

        commands = func_dict[func_name]['values']
        print(' ', f'COMMAND: {commands[0]}, {commands[1]}')

        if func_dict[func_name]['args'] is not None:
            for arg_name in func_dict[func_name]['args']:
                aliases = arg_dict[arg_name]['values']

                print(
                    ' '*indent, f'OPTION: {aliases[0]}, {aliases[1]}, {aliases[2]}, {aliases[3]}')

                if arg_dict[arg_name]['required']:
                    print(' '*2*indent, 'REQUIRED')

                print(' '*2*indent, f'NAME: {arg_dict[arg_name]["name"]}')

                if arg_dict[arg_name]['default'] is not None:
                    print(' '*2*indent,
                          f'DEFAULT: {arg_dict[arg_name]["default"]}')

                if arg_dict[arg_name]['syntax'] is not None:
                    print(' '*2*indent,
                          f'FORMAT: {arg_dict[arg_name]["syntax"]}')
        separator_line()

# ANALYSIS SPECIFIC OUTPUT FUNCTIONS


def portfolio_percent_result(result, tickers, indent=formats.formats['INDENT']):
    for i, item in enumerate(tickers):
        print(' '*indent, f'{item} =', round(100*result[i], 2), '%')


def portfolio_shares_result(result, tickers, indent=formats.formats['INDENT']):
    for i, item in enumerate(tickers):
        print(' '*indent, f'{item} =', result[i])


def spot_price(ticker, this_spot_price):
    formatted_price = round(float(this_spot_price), 2)
    scalar_result(f'{ticker} spot price', formatted_price)


def model_price(ticker: str, this_model_price: Union[str, float], model: str) -> None:
    formatted_price = round(float(this_model_price), 2)
    scalar_result(f'{ticker} {str(model).upper()} price', formatted_price)


def risk_profile(profiles: Dict[str, Dict[str, float]]) -> None:
    for key, value in profiles.items():
        title_line(f'{key} Risk Profile')
        for subkey, subvalue in value.items():
            scalar_result(f'{subkey}', f'{subvalue}', currency=False)


def moving_average_result(ticker: str, averages: Dict[str, Dict[str, float]]) -> None:
    """
    Prints the results of `scrilla.analysis.models.geometric.statistics.calculate_moving_averages` or `scrilla.analysis.models.reversion.statistics.calculate_moving_averages` to *stdout*.

    Parameters
    ----------
    1. **averages**: ``Dict[str, Dict[str,float]]``
        The dictionary returned from a call to `scrilla.analysis.models.geometric.statistics.calculate_moving_averages` or `scrilla.analysis.models.reversion.statistics.calculate_moving_averages`.
    """
    title_line(f'{ticker} Moving Averages')
    for this_date, average_dict in averages.items():
        center(this_date)
        for avg_key, average in average_dict.items():
            scalar_result(calculation=avg_key, result=average, currency=False)


def screen_results(info, model):
    for ticker in info:
        title_line(f'{ticker} {str(model).upper()} Model vs. Spot Price ')
        spot_price(ticker=ticker, this_spot_price=info[ticker]['spot_price'])
        model_price(ticker=ticker,
                    this_model_price=info[ticker]['model_price'], model=model)
        scalar_result(f'{ticker} discount', info[ticker]['discount'])
        separator_line()

# TODO: can probably combine optimal_result and efficient_frontier into a single function
#         by wrapping the optimal_results in an array so when it iterates through frontier
#         in efficient_frontier, it will only pick up the single allocation array for the
#         optimal result.


def optimal_result(portfolio, allocation, investment=None, latest_prices=None):
    title_line('Optimal Percentage Allocation')
    portfolio_percent_result(allocation, portfolio.tickers)

    if investment is not None:
        shares = portfolio.calculate_approximate_shares(
            allocation, investment, latest_prices)
        total = portfolio.calculate_actual_total(
            allocation, investment, latest_prices)

        title_line('Optimal Share Allocation')
        portfolio_shares_result(shares, portfolio.tickers)
        title_line('Optimal Portfolio Value')
        scalar_result('Total', round(total, 2))

    title_line('Risk-Return Profile')
    scalar_result(calculation='Return', result=portfolio.return_function(
        allocation), currency=False)
    scalar_result(calculation='Volatility', result=portfolio.volatility_function(
        allocation), currency=False)


def efficient_frontier(portfolio, frontier, investment=None, latest_prices=None):
    title_line('(Annual Return %, Annual Volatility %) Portfolio')

    # TODO: edit title to include dates

    for allocation in frontier:
        separator_line()
        return_string = str(
            round(round(portfolio.return_function(allocation), 4)*100, 2))
        vol_string = str(
            round(round(portfolio.volatility_function(allocation), 4)*100, 2))
        title_line(f'({return_string} %, {vol_string}%) Portfolio')
        separator_line()

        title_line('Optimal Percentage Allocation')
        portfolio_percent_result(allocation, portfolio.tickers)

        if investment is not None:
            shares = portfolio.calculate_approximate_shares(
                allocation, investment, latest_prices)
            total = portfolio.calculate_actual_total(
                allocation, investment, latest_prices)

            title_line('Optimal Share Allocation')
            portfolio_shares_result(shares, portfolio.tickers)
            title_line('Optimal Portfolio Value')
            scalar_result('Total', round(total, 2))

        title_line('Risk-Return Profile')
        scalar_result('Return', portfolio.return_function(
            allocation), currency=False)
        scalar_result('Volatility', portfolio.volatility_function(
            allocation), currency=False)
        print('\n')


def correlation_matrix(tickers: List[str], correl_matrix: List[List[float]], display: bool = True):
    """
    Parameters
    ----------
    1. **tickers**: ``list``
        Array of tickers for which the correlation matrix was calculated and formatted.
    2. **indent**: ``int``
        Amount of indent on each new line of the correlation matrix.
    3. **start_date**: ``datetime.date`` 
        Start date of the time period over which correlation was calculated. 
    4. **end_date**: ``datetime.date`` 
        End date of the time period over which correlation was calculated. 

    Returns
    ------
    A correlation matrix string formatted with new lines and spaces.
    """
    entire_formatted_result, formatted_subtitle, formatted_title = "", "", ""

    line_length, first_symbol_length = 0, 0
    new_line = ""
    no_symbols = len(tickers)

    for i in range(no_symbols):
        this_symbol = tickers[i]
        symbol_string = ' '*formats.formats['INDENT'] + f'{this_symbol} '

        if i != 0:
            this_line = symbol_string + ' ' * \
                (line_length - len(symbol_string) - 7*(no_symbols - i))
            # NOTE: seven is number of chars in ' 100.0%'
        else:
            this_line = symbol_string
            first_symbol_length = len(this_symbol)

        new_line = this_line

        for j in range(i, no_symbols):
            if j == i:
                new_line += " 100.0%"

            else:
                result = correl_matrix[i][j]
                formatted_result = str(
                    100*result)[:constants.constants['SIG_FIGS']]
                new_line += f' {formatted_result}%'

        entire_formatted_result += new_line + '\n'

        if i == 0:
            line_length = len(new_line)

    formatted_subtitle += ' ' * \
        (formats.formats['INDENT'] + first_symbol_length+1)
    for i, ticker in enumerate(tickers):
        sym_len = len(ticker)
        formatted_subtitle += f' {ticker}' + ' '*(7-sym_len)
        # NOTE: seven is number of chars in ' 100.0%'
        if i == 0:
            formatted_title += f'({ticker},'
        elif i < len(tickers)-1:
            formatted_title += f'{ticker},'
        else:
            formatted_title += f'{ticker}) correlation matrix'

    formatted_subtitle += '\n'

    whole_thing = formatted_subtitle + entire_formatted_result

    if display:
        title_line(formatted_title)
        print(f'\n{whole_thing}')
        return

    return whole_thing


class Logger():

    def __init__(self, location, log_level="info"):
        self.location = location
        self.log_level = log_level

    # LOGGING FUNCTIONS
    def comment(self, msg):
        now = datetime.datetime.now()
        dt_string = now.strftime("%d/%m/%Y %H:%M:%S")
        print(dt_string, ' :', self.location, ' : ', msg)

    def info(self, msg):
        if self.log_level in [constants.constants['LOG_LEVEL']['INFO'],
                              constants.constants['LOG_LEVEL']['DEBUG'],
                              constants.constants['LOG_LEVEL']['VERBOSE']]:
            self.comment(msg)

    def debug(self, msg):
        if self.log_level in [constants.constants['LOG_LEVEL']['DEBUG'],
                              constants.constants['LOG_LEVEL']['VERBOSE']]:
            self.comment(msg)

    def verbose(self, msg):
        if self.log_level == constants.constants['LOG_LEVEL']['VERBOSE']:
            self.comment(msg)



import json

from typing import Tuple

from scrilla import settings
from scrilla.util import helper
from scrilla.static import keys

MARGINS = 5


def format_stylesheet(sheet):
    theme = get_mode_theme()
    for element in theme:
        sheet = sheet.replace(element, theme[element])

    with open(settings.GUI_ICON_FILE, 'r') as f:
        ICONS = json.load(f)
        for icon in ICONS:
            for state in ICONS[icon]:
                sheet = sheet.replace(
                    f'{icon}-{state}', f'{settings.ASSET_DIR}/{ICONS[icon][state]}')
    return sheet


def get_mode_theme():
    with open(settings.GUI_THEME_FILE, 'r') as f:
        MATERIAL = json.load(f)

    if settings.GUI_DARK_MODE:
        theme_key = 'dark_mode'
    else:
        theme_key = 'light_mode'

    theme = {}
    for i, color in enumerate(MATERIAL[theme_key]):
        for scheme in MATERIAL[color]:
            if i == 0:
                theme[f'$primary-{scheme}'] = MATERIAL[color][scheme]
            elif i == 1:
                theme[f'$accent-{scheme}'] = MATERIAL[color][scheme]
            elif i == 2:
                theme[f'$warn-{scheme}'] = MATERIAL[color][scheme]
    return theme


def get_light_mode_theme():
    with open(settings.GUI_THEME_FILE, 'r') as f:
        MATERIAL = json.load(f)

    light_theme = {}
    for color in MATERIAL['grey']:
        light_theme[f'$primary-{color}'] = MATERIAL['grey'][color]
    for color in MATERIAL['green']:
        light_theme[f'$accent-{color}'] = MATERIAL['green'][color]
    for color in MATERIAL['red']:
        light_theme[f'$warn-{color}'] = MATERIAL['red'][color]

    return light_theme


def format_allocation_profile_title(allocation, portfolio) -> str:
    port_return, port_volatility = portfolio.return_function(
        allocation), portfolio.volatility_function(allocation)
    formatted_result = "("+str(100 *
                               port_return)[:5]+"%, " + str(100*port_volatility)[:5]+"%)"
    formatted_result_title = "("
    for symbol in portfolio.tickers:
        if portfolio.tickers.index(symbol) != (len(portfolio.tickers) - 1):
            formatted_result_title += symbol+", "
        else:
            formatted_result_title += symbol + ") Portfolio Return-Risk Profile"
    whole_thing = formatted_result_title + " = "+formatted_result
    return whole_thing


def format_profile(profile: dict) -> Tuple[str]:
    profile_keys = keys.keys['APP']['PROFILE']
    for key in profile_keys:
        if key in ['RET', 'VOL', 'EQUITY']:
            profile = helper.format_dict_percent(profile, profile_keys[key])
        else:
            profile = helper.format_dict_number(profile, profile_keys[key])
    return profile
